'use strict';

const redis = require('redis');
const bluebird = require('bluebird');
const redisActionsMap = require('./consts/redisActionsMap');
const { CACHE_STAR_PREFIX, CS_PREFIX_OFFSET, NOT_REDIS_VALUE, GET, SET, DEL, GET_ALL } = require('./consts/consts');
bluebird.promisifyAll(redis.RedisClient.prototype);
bluebird.promisifyAll(redis.Multi.prototype);

exports = module.exports = cacheStar;

function cacheStar(redisObj){
    this.redisClient = redis.createClient(redisObj);
    this.redisClientSubscriber = redis.createClient(redisObj);
    this.cache = {};
    this.subscribeObj = {};
    this.redisClientSubscriber.on('message', (channel, keyMsg) => {
        channel = channel.substring(CS_PREFIX_OFFSET);
        if(keyMsg){
            channel = channel.concat(`.${keyMsg}`).split('.');
        }

        let dataFromCache = this.getCacheObj(channel);
        this.removeCacheObj(dataFromCache);
        this.runRedisCommand(GET, ...channel)
        .then((dataFromRedis) => {
            Object.assign(dataFromCache, dataFromRedis);
        });
    });
}

cacheStar.prototype.getCacheObj = function(objPathArr){
    let res = objPathArr.reduce((accCache, prop) => {
        if(!(prop in accCache)){
            accCache[prop] = {};
        }
        return accCache[prop];
    }, this.cache);

    return res;
};

cacheStar.prototype.getAll = function(...objPathArr){
    return new Promise((resolve, reject) => {
        let dataFromCache = this.getCacheObj(objPathArr);
        let [objPath] = objPathArr;
        const CHANNEL = `${CACHE_STAR_PREFIX}:${objPath}`;

        if (dataFromCache === null) {
            dataFromCache = {};
        }

        if(Object.keys(dataFromCache).length){
            return resolve(Object.assign({}, dataFromCache));
        }

        if(!this.subscribeObj[CHANNEL]) {
            this.subscribeObj[CHANNEL] = true;
            this.subscribe(CHANNEL);
        }

        this.runRedisCommand(GET_ALL, ...objPathArr).then((dataFromRedis) => {
            if (dataFromRedis === null) {
                dataFromRedis = {};
            }

            if(!Object.keys(dataFromRedis).length){
                this.removeCachePath(objPathArr);
            }

            Object.assign(dataFromCache, dataFromRedis);
            return resolve(dataFromCache);
        });
    }).catch((err) => {reject(err);});

};

cacheStar.prototype.removeCachePath = function(removeCachePathArr){
    let pathToRemove = this.cache;
    for (let prop of removeCachePathArr) {
        
        if(pathToRemove && pathToRemove[prop] && !Object.keys(pathToRemove[prop]).length){
            delete pathToRemove[prop];
        }

        pathToRemove = pathToRemove && pathToRemove[prop] ? pathToRemove[prop]: null;
        if (!pathToRemove) break;
    }

};

cacheStar.prototype.get = function(...objPathArr){
    return new Promise((resolve, reject) => {
        let dataFromCache = this.getCacheObj(objPathArr);
        let objPath = objPathArr.slice(0, -1).join('.') ? objPathArr.slice(0, -1).join('.') : objPathArr.slice().join('.');
        const CHANNEL = `${CACHE_STAR_PREFIX}:${objPath}`;

        if(dataFromCache === null) {
            dataFromCache = {};
        }

        if(Object.keys(dataFromCache).length){
            return resolve(Object.assign({}, dataFromCache));
        }

        if(!this.subscribeObj[CHANNEL]) {
            this.subscribeObj[CHANNEL] = true;
            this.subscribe(CHANNEL);
        }

        this.runRedisCommand(GET, ...objPathArr).then((dataFromRedis) => {
            if (dataFromRedis === null) {
                dataFromRedis = {};
            }

            if(!Object.keys(dataFromRedis).length){
                this.removeCachePath(objPathArr);
            }

            Object.assign(dataFromCache, dataFromRedis);
            return resolve(dataFromCache);
        });

    }).catch((err) => {reject(err)})
};


cacheStar.prototype.del = function(...objPathArr){
    return new Promise((resolve, reject) => {
        this.runRedisCommand(DEL, ...objPathArr).then(() => {
            return resolve(this.publish(...objPathArr));
        }).catch(err => reject(err));

    });
};

cacheStar.prototype.runRedisCommand = function(method, ...objPathArr){
    return new Promise((resolve, reject) => {
        this.redisClient[redisActionsMap[objPathArr.length][method]](...objPathArr)
        .then((ans) => {
            if (method == GET_ALL) {
                let parsedObj = {};
                if(ans === null) {
                    ans = {};
                }

                Object.keys(ans).map((campaignKey) => {
                    parsedObj[campaignKey] = JSON.parse(ans[campaignKey]);
                });

                return resolve(parsedObj);
            }
            if(Array.isArray(ans)){
                return resolve(ans);
            }else{
                return resolve(JSON.parse(ans));
            }
        }).catch((err) => {reject(err)});
    });
};

cacheStar.prototype.removeCacheObj = function(obj){
    if(obj === null) {
        obj = {};
    }

    Object.keys(obj).forEach((prop) => {
        delete obj[prop];
    });
};

cacheStar.prototype.saveInRedis = function(objToSave, ...objPathArr){
    return this.redisClient[redisActionsMap[objPathArr.length][SET]](...objPathArr, objToSave);
};

cacheStar.prototype.subscribe = function(CHANNEL){
    return this.redisClientSubscriber.subscribe(CHANNEL);
};

cacheStar.prototype.publish = function(...args){
    let lastKey;
    let objPath;
    let CHANNEL;

    if(args.length > 1) {
        lastKey = args.pop();
        objPath = args.join('.');
        CHANNEL = `${CACHE_STAR_PREFIX}:${objPath}`;
        return this.redisClient.publish(CHANNEL, lastKey);
    }else{
        [objPath] = args;
        CHANNEL = `${CACHE_STAR_PREFIX}:${objPath}`;
        return this.redisClient.publish(CHANNEL);
    }


};

cacheStar.prototype.set = function(...args){
    return new Promise((resolve, reject) => {
        let objToSave = args.pop();

        this.saveInRedis(objToSave, ...args).then(() => {
            return resolve(this.publish(...args));
        });
    }).catch(err => reject(err));
};
